/*
    $Id: average.h 143 2010-04-04 08:42:34Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _AVERAGE_H
#define _AVERAGE_H

#include <cmath>
#include <iterator>

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/io.hpp>

#include "apply_to_all.h"

namespace utils {
namespace algo {

	template<class T> struct mean;
	template<class T> struct variance;

	template<class T> struct mean {
	public:
		typedef T value_type;
		typedef T result_type;

		void operator() ( result_type& result, const value_type& x ) const {
			result += x;
		}
		result_type initial( const value_type& x ) const {
			return result_type( x );
		}
		template<class S> void result( result_type& result, S size ) const {
			result /= size;
		}
	private:
	};

	struct variance_traits {
		static double prod( double x, double y ){
			return x*y;
		}
		static float prod( float x, float y ){
			return x*y;
		}
		template< class V1, class V2 > static
		typename boost::numeric::ublas::vector_binary_traits< V1, V2, boost::numeric::ublas::scalar_multiplies< typename V1::value_type, typename V2::value_type > >::result_type
		prod( const boost::numeric::ublas::vector_expression< V1 >& e1, const boost::numeric::ublas::vector_expression< V2 >& e2 ) {
			return element_prod( e1, e2 );
		}
		template< class M1, class M2 > static
		typename boost::numeric::ublas::matrix_binary_traits< M1, M2, boost::numeric::ublas::scalar_multiplies< typename M1::value_type, typename M2::value_type > >::result_type
		prod( const boost::numeric::ublas::matrix_expression< M1 >& e1, const boost::numeric::ublas::matrix_expression< M2 >& e2 ) {
			return element_prod( e1, e2 );
		}

		static double sqrt( double x ){
			return std::sqrt(x);
		}
		static float sqrt( float x ){
			return std::sqrt(x);
		}
		template< class V1 > static
		typename boost::numeric::ublas::vector_unary_traits< V1, utils::ublas::detail::sqrt< typename V1::value_type > >::result_type
		sqrt( const boost::numeric::ublas::vector_expression< V1 >& e1) {
			return utils::ublas::sqrt( e1 );
		}
		template< class M1 > static
		typename boost::numeric::ublas::matrix_unary1_traits< M1, utils::ublas::detail::sqrt< typename M1::value_type > >::result_type
		sqrt( const boost::numeric::ublas::matrix_expression< M1 >& e1) {
			return utils::ublas::sqrt( e1 );
		}
	};

	struct covariance_traits {
		template< class V1, class V2 > static
		boost::numeric::ublas::symmetric_adaptor< const typename boost::numeric::ublas::vector_matrix_binary_traits<V1, V2, boost::numeric::ublas::scalar_multiplies<typename V1::value_type, typename V2::value_type> >::result_type, boost::numeric::ublas::upper >
		prod( const boost::numeric::ublas::vector_expression< V1 >& e1, const boost::numeric::ublas::vector_expression< V2 >& e2 ) {
			typedef typename boost::numeric::ublas::vector_matrix_binary_traits<V1, V2, boost::numeric::ublas::scalar_multiplies<typename V1::value_type, typename V2::value_type> >::result_type outer_prod_type;
			return boost::numeric::ublas::symmetric_adaptor< const outer_prod_type, boost::numeric::ublas::upper >( outer_prod( e1, e2 ) );
		}
	};
	
	template<class T> struct variance {
	public:
		typedef T value_type;
		typedef T result_type;
		
		variance( const value_type& mean ): mean_(mean) {}
		
		void operator() ( result_type& result, const value_type& x ) const {
			result += variance_traits::prod( x - mean_, x - mean_ );
		}
		result_type initial( const value_type& x ) const {
			return result_type( variance_traits::prod( x - mean_, x - mean_ ) );
		}
		template<class S> void result( result_type& result, S size ) const {
			result /= size-1;
		}
	private:
		value_type mean_;
	};

	template<class T> struct covariance {
	public:
		typedef T value_type;
		typedef boost::numeric::ublas::symmetric_matrix< typename value_type::value_type > result_type;

		covariance( const value_type& mean ): mean_(mean) {}

		void operator() ( result_type& result, const value_type& x ) const {
			result += covariance_traits::prod( x - mean_, x - mean_ );
		}
		result_type initial( const value_type& x ) const {
			return result_type( covariance_traits::prod( x - mean_, x - mean_ ) );
		}
		template<class S> void result( result_type& result, S size ) const {
			result /= size-1;
		}
	private:
		value_type mean_;
	};

	template<class T> struct error_of_mean: variance<T> {
		error_of_mean( const typename variance<T>::value_type& mean ): variance<T>(mean) {}
		template<class S> void result( typename variance<T>::result_type& result, S size ) const {
			variance<T>::result( result, size );
			result = variance_traits::sqrt( result / size );
		}
	};
	
	template< class Iterator, template<class T> class F >
	typename F< typename std::iterator_traits<Iterator>::value_type >::result_type
	average( Iterator begin, Iterator end, F< typename std::iterator_traits<Iterator>::value_type > op ) {
		typedef Iterator iterator;
		typedef typename std::iterator_traits< iterator >::value_type value_type;
		typedef typename std::iterator_traits< iterator >::difference_type difference_type;
		typedef difference_type size_type;
		typedef F< value_type > operator_type;
		typedef typename operator_type::result_type result_type;

		if( begin == end )
			return result_type();
		
		result_type ret( op.initial( *begin ) );
		size_type n = 1;
		
		for( iterator it = ++begin; it != end; ++it,++n ) {
			op( ret, *it );
		}

		op.result( ret, n );
		return ret;
	}

} // algo
} // utils

#endif // _AVERAGE_H
