/*
    $Id: background_factory.h 147 2010-04-17 09:25:25Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _BACKGROUND_FACTORY_H
#define _BACKGROUND_FACTORY_H

#include <list>
#include <memory>

#include <boost/function.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

using namespace boost::numeric;

/* Forward declaration */
class background_factory;

class background_model;
class background_daytime_model;
class background_twilight_model;
class background_night_model;

class background_abstract_filter;
class background_nonpoisson_filter;

/* Declaration */
class background_factory {
	public:
		typedef background_abstract_filter filter_type;
		typedef background_model  model_type;

		typedef boost::function1<model_type*, const boost::posix_time::ptime&> select_model_function_type;

		typedef double                     value_type;
		typedef ublas::vector<value_type>  vector_type;
		typedef vector_type::size_type     size_type;

		typedef vector_type flux_type;
		typedef ublas::symmetric_matrix<value_type> vars_type;
	private:
		std::auto_ptr< model_type > m_curmodel;
		select_model_function_type  m_select_model;
	public:
		background_factory(const select_model_function_type& select_model);
		background_factory(const background_factory& factory);
		void swap(background_factory& factory) throw();
		background_factory& operator=(const background_factory& factory);
		~background_factory();

		vector_type background( const boost::posix_time::ptime& time ) const;

		model_type* create_new_model( const boost::posix_time::ptime& time ) const;
		model_type* release_current_model();
		void handle_new_model( model_type* model );
};

class background_model {
	public:
		typedef background_factory::value_type  value_type;
		typedef background_factory::vector_type vector_type;
		typedef background_factory::size_type   size_type;
	private:
		boost::posix_time::ptime m_inittime;
	public:
		background_model( const boost::posix_time::ptime& time );
		virtual ~background_model() = 0;

		const boost::posix_time::ptime& inittime() const { return m_inittime; }

		virtual size_type size() const = 0;
		virtual vector_type background( const boost::posix_time::ptime& time ) const = 0;
		virtual void accumulate( const boost::posix_time::ptime& time, const vector_type& mean ) = 0;

		virtual background_model* clone() const = 0;
};

class background_daytime_model: public background_model {
	private:
		size_type   m_size;
	public:
		background_daytime_model( const boost::posix_time::ptime& time );

		size_type size() const;
		vector_type background( const boost::posix_time::ptime& time ) const;
		void accumulate( const boost::posix_time::ptime& time, const vector_type& mean );

		background_model* clone() const;
};

class background_twilight_model: public background_model {
	public:
		typedef boost::function1<value_type, const boost::posix_time::ptime&> sun_altitude_function_type;
	private:
		size_type   m_size;
		vector_type m_background;
		sun_altitude_function_type m_sun_altitude;
	private:
		value_type extrapolate( const boost::posix_time::ptime& timef ) const;
	public:
		background_twilight_model( const boost::posix_time::ptime& time, const sun_altitude_function_type& sun_altitude );

		size_type size() const;
		vector_type background( const boost::posix_time::ptime& time ) const;
		void accumulate( const boost::posix_time::ptime& time, const vector_type& mean );

		background_model* clone() const;
};

class background_night_model: public background_model {
	private:
		size_type   m_size;
		vector_type m_background;
	public:
		background_night_model( const boost::posix_time::ptime& time );

		size_type size() const;
		vector_type background( const boost::posix_time::ptime& time ) const;
		void accumulate( const boost::posix_time::ptime& time, const vector_type& mean );

		background_model* clone() const;
};

class background_nonpoisson_filter {
	public:
		typedef background_factory::value_type  value_type;
		typedef background_factory::vector_type vector_type;
		typedef background_factory::size_type   size_type;

		typedef background_factory::flux_type   flux_type;
		typedef background_factory::vars_type   vars_type;
	private:
		value_type m_threshold;
		value_type m_nonpoissonD;
		value_type m_numexp;
	public:
		background_nonpoisson_filter( const value_type& threshold, const value_type& nonpoissonD, const value_type& numexp );
		~background_nonpoisson_filter();

		bool operator() ( const boost::posix_time::ptime& time, const flux_type& flux, const vars_type& vars ) const;
};

#endif // _BACKGROUND_FACTORY_H
