/*
    $Id: dimm.h 166 2010-10-29 15:22:58Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _DIMM_H
#define _DIMM_H

#include "table_storage.h"

#include <boost/function.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/numeric/ublas/vector.hpp>

namespace atmos {
namespace dimm {
	using namespace boost::numeric::ublas;

	typedef boost::posix_time::ptime time_type;
	typedef double                   value_type;
	typedef vector<value_type>       vector_type;
	typedef vector_type::size_type   size_type;

	typedef vector_type  flux_type;
	typedef vector_type  vars_type;
	typedef unsigned int franenum_type;

	vars_type  correction( const vars_type& vars );
	value_type turbulence( value_type scale, value_type Knumber, value_type variance );
	value_type seeing( value_type turbulence_power );

namespace detail {
	struct sample {
		typedef atmos::dimm::time_type     timestamp_type;
		typedef atmos::dimm::franenum_type framenum_type;
		typedef atmos::dimm::flux_type     flux_type;
		typedef atmos::dimm::vars_type     vars_type;
		
		timestamp_type timestamp;
		framenum_type  framenum;
		flux_type      left_flux;
		flux_type      right_flux;
		vars_type      xvars;
		vars_type      yvars;

		sample( const timestamp_type& time,
			     const framenum_type& framenum,
			     const flux_type& left,
			     const flux_type& right,
			     const vars_type& xvars,
			     const vars_type& yvars ):
			timestamp(time), framenum(framenum), left_flux(left), right_flux(right), xvars(xvars), yvars(yvars) {}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::flux_type, detail::sample > > left_flux_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::flux_type, detail::sample > > right_flux_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::vars_type, detail::sample > > xvars_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::vars_type, detail::sample > > yvars_column_type;
private:
	left_flux_column_type  m_left_flux_column;
	right_flux_column_type m_right_flux_column;
	xvars_column_type      m_xvars_column;
	yvars_column_type      m_yvars_column;
public:
	storage( size_type reserve = 60 );

	const left_flux_column_type&  left_flux()  const { return m_left_flux_column; }
	const right_flux_column_type& right_flux() const { return m_right_flux_column; }
	const xvars_column_type&      xvars()      const { return m_xvars_column; }
	const yvars_column_type&      yvars()      const { return m_yvars_column; }

	void push_back( const time_type& time,
			const franenum_type& framenum,
			const flux_type& left,
			const flux_type& right,
			const vars_type& xvars,
			const vars_type& yvars ) {

		table_storage< detail::sample >::push_back( detail::sample( time, framenum, left, right, xvars, yvars ) );
	}

	void correct_low_vars();
};

}}

#endif // _DIMM_H
