/*
    $Id: file_output.cpp 179 2010-12-16 21:09:18Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <boost/io/ios_state.hpp>

#include "file_output.h"
#include "integrals.h"

file_output::file_output( std::ostream& stm ): m_stm(stm) {
	boost::posix_time::time_facet* facet(new boost::posix_time::time_facet("%Y-%m-%d %H:%M:%S"));
	stm.imbue(std::locale(stm.getloc(), facet));
}
file_output::~file_output() {

}
void file_output::comment( const std::string& comment ) {
	m_stm << "# " << comment << std::endl;
}
void file_output::normal( const time_type& time, double mz ) {
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "M " << time << " Normal " << std::fixed << std::setprecision(3) << mz << std::endl;
}
void file_output::background( const time_type& time ) {
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "M " << time << " Background measurements" << std::endl;
}
void file_output::atmosphere( const time_type& time, double free_seeing, double free_seeing_err, double dimm_seeing, double dimm_seeing_err, double dimm_turbulence, double dimm_turbulence_err, double isoplanatic_angle, double isoplanatic_angle_err, double altitude_eff, double altitude_eff_err, double time_constant, double time_constant_err, double moment_0, double moment_0_err, double moment_2, double moment_2_err ) {
	static datum skipped;
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "A " << time << " "
		<< std::fixed << datum( free_seeing, free_seeing_err/free_seeing ) << " " << datum( dimm_seeing, dimm_seeing_err/dimm_seeing ) << " "
		<< std::scientific << datum( moment_0, moment_0_err/moment_0 ) << " " << datum( dimm_turbulence, dimm_turbulence_err/dimm_turbulence ) << " "
		<< std::fixed << datum( altitude_eff, altitude_eff_err/altitude_eff ) << " " << skipped << " "
		<< datum( isoplanatic_angle, isoplanatic_angle_err/isoplanatic_angle ) << " "
		<< std::scientific << datum( moment_2, moment_2_err/moment_2 ) << " "
		<< std::fixed << datum( time_constant, time_constant_err/time_constant ) << std::endl;
}
void file_output::profile( const time_type& time, const std::string& method, double cn2seeing, double chi2, const ublas::vector<double>& cn2, const ublas::vector<double>& ecn2, const ublas::vector<double>& grid ){
	assert( cn2.size() == ecn2.size() );
	assert( cn2.size() == grid.size() );

	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "T " << time << " " << method << " "
		<< std::setw(2) << cn2.size() << " " << std::setprecision(2)
		<< std::fixed << chi2 << " " << cn2seeing;
	for( ublas::vector<double>::size_type i = 0; i < cn2.size(); i++)
		m_stm << " " << std::fixed << grid(i)
			<< std::scientific  << " " << cn2(i) << " " << ecn2(i);
	m_stm << std::endl;
}
void file_output::residual( const time_type& time, const std::string& method, const ublas::vector<double>& o_csi ) {
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "R " << time << " " << method << " "
		<< std::setprecision(3) << std::fixed;
	for( ublas::vector<double>::size_type i = 0; i < o_csi.size(); i++ )
		m_stm << "  " << o_csi(i);
	m_stm << std::endl;
}
void file_output::fluxes( const time_type& time, const ublas::vector<double>& result_flux ) {
	boost::io::ios_all_saver asaver( m_stm );
	assert( result_flux.size() >= 4 );
	m_stm << "F " << time << " " << std::fixed << datum( result_flux(0) ) << ' ' << datum( result_flux(1) ) << ' ' << datum( result_flux(2) ) << ' ' << datum( result_flux(3) ) << std::endl;
}
void file_output::fluxes( const time_type& time, const ublas::vector<double>& result_flux, const ublas::vector<double>& result_flux_err ) {
	boost::io::ios_all_saver asaver( m_stm );
	assert( result_flux.size() >= 4 );
	assert( result_flux.size() == result_flux_err.size() );
	m_stm << "F " << time << " " << std::fixed
	<< datum( result_flux(0), result_flux_err(0)/result_flux(0) ) << ' '
	<< datum( result_flux(1), result_flux_err(1)/result_flux(1) ) << ' '
	<< datum( result_flux(2), result_flux_err(2)/result_flux(2) ) << ' '
	<< datum( result_flux(3), result_flux_err(3)/result_flux(3) ) << std::endl;
}
void file_output::indices( const time_type& time, const ublas::symmetric_matrix<double>& result_indx, const ublas::symmetric_matrix<double>& result_indx_err, const ublas::vector<double>& result_desi, const ublas::vector<double>& result_desi_err ) {
	boost::io::ios_all_saver asaver( m_stm );
	ios_indic_saver isaver( m_stm );
	assert( result_desi.size() >= 4 );
	assert( result_indx.size1() >= 4 );
	assert( result_indx.size2() >= 4 );
	assert( result_desi.size() == result_desi_err.size() );
	assert( result_indx.size1() == result_indx_err.size1() );
	assert( result_indx.size1() == result_indx_err.size2() );
	m_stm << "I " << time << std::indic << std::fixed <<
		datum( result_indx(0,0), result_indx_err(0,0)/result_indx(0,0) ) << datum( result_indx(1,1), result_indx_err(1,1)/result_indx(1,1) ) << datum( result_indx(2,2), result_indx_err(2,2)/result_indx(2,2) ) << datum( result_indx(3,3), result_indx_err(3,3)/result_indx(3,3) ) <<
		datum( result_indx(0,1), result_indx_err(0,1)/result_indx(0,1) ) << datum( result_indx(0,2), result_indx_err(0,2)/result_indx(0,2) ) << datum( result_indx(0,3), result_indx_err(0,3)/result_indx(0,3) ) <<
		datum( result_indx(1,2), result_indx_err(1,2)/result_indx(1,2) ) << datum( result_indx(1,3), result_indx_err(1,3)/result_indx(1,3) ) <<
		datum( result_indx(2,3), result_indx_err(2,3)/result_indx(2,3) ) <<
		datum( result_desi(0), result_desi_err(0)/result_desi(0) ) << datum( result_desi(1), result_desi_err(1)/result_desi(1) ) << datum( result_desi(2), result_desi_err(2)/result_desi(2) ) << datum( result_desi(3), result_desi_err(3)/result_desi(3) ) << std::endl;
}
void file_output::parameter( const time_type& time, const std::string& key, const std::string& value ) {
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "P " << time << " " << key << " = " << value << std::endl;
}
void file_output::objectinfo( const time_type& time, const object_type& obj ){
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "O " << time << " " << obj << std::endl;
}
void file_output::wind( const time_type& time, const std::string& mode, double chi2, double tau0, const ublas::vector<double>& profile, const ublas::vector<double>& eps, const ublas::vector<double>& grid ) {
	assert( profile.size() == eps.size() );
	
	boost::io::ios_all_saver asaver( m_stm );
	m_stm << "W " << time << " " << mode << " "
		<< std::setw(2) << profile.size() << " " << std::setprecision(2)
		<< std::fixed << chi2 << " " << tau0;
	for( ublas::vector<double>::size_type i = 0; i < profile.size(); i++)
		m_stm << " " << std::fixed << grid(i)
			<< std::scientific << " " << profile(i) << " "
			<< std::setprecision(2) << std::fixed << ( std::isfinite(eps(i)) ? eps(i) : 0.0 );
	m_stm << std::endl;
}
