/*
    $Id: indices.cpp 177 2010-12-13 10:37:30Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <numeric>

#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include "atmos.h"
#include "indices.h"
#include "io.h"
#include "utils.h"

#include "photometry.h"

namespace atmos {
namespace indices {

flux_type flux_nonlinearity_correction ( const flux_type& flux, const vector_type& deadtime ) {
	return photometry::flux_nonlinearity_correction( flux, deadtime );
}
flux_type flux_background_correction( const flux_type& flux, const vector_type& scatter, const vector_type& background ) {
	return photometry::flux_scatter_correction( flux, scatter ) - background;
}
boost::tuple< vars_type, rho1_type, rho2_type > vars_nonlinearity_correction( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2, const vector_type& deadtime, const vector_type& npoisson, const value_type& exposition ) {
	using namespace utils::ublas;

	typedef matrix_vector_slice< vars_type > diagonal_type;
	typedef boost::tuple< vars_type, rho1_type, rho2_type > result_type;

	assert( vars.size1() == flux.size() );
	assert( vars.size2() == flux.size() );
	assert( rho1.size1() == flux.size() );
	assert( rho1.size2() == flux.size() );
	assert( rho2.size() == flux.size() );
	
	result_type ret(vars,rho1,rho2);

	diagonal_type variances = diag( ret.get<0>() );
	variances -= photometry::variance_nonpoisson_correction( flux, deadtime, npoisson ) / exposition;
	
	flux_type corr = photometry::variance_nonlinearity_correction( flux, deadtime );

	ret.get<0>() = element_div( ret.get<0>(), outer_prod( corr, corr ) );
	ret.get<1>() = element_div( ret.get<1>(), outer_prod( corr, corr ) );
	ret.get<2>() = element_div( ret.get<2>(), element_prod( corr, corr ) );

	return ret;
}
vars_type vars_nonlinearity_correction( const flux_type& flux, const vars_type& vars, const vector_type& deadtime, const vector_type& npoisson, const value_type& exposition ) {
	using namespace utils::ublas;

	typedef matrix_vector_slice< vars_type > diagonal_type;

	assert( vars.size1() == flux.size() );
	assert( vars.size2() == flux.size() );

	vars_type ret( vars );

	diagonal_type variances = diag( ret );
	variances -= photometry::variance_nonpoisson_correction( flux, deadtime, npoisson ) / exposition;

	flux_type corr = photometry::variance_nonlinearity_correction( flux, deadtime );

	return element_div( ret, outer_prod( corr, corr ) );
}

vars_type vars_exposure_correction( const flux_type& flux, const vars_type& vars, const rho1_type& rho1 ) {
	return vars*1.25 - rho1*0.25;
}

indices_type calculate_indices( const flux_type& flux, const vars_type& vars ) {
	using namespace utils::ublas;

	typedef matrix_vector_slice< indices_type >    indices_diagonal_type;

	assert( vars.size1() == flux.size() );
	assert( vars.size2() == flux.size() );

	indices_type ret = utils::ublas::log( xp1( element_div( vars, outer_prod( flux, flux ) ) ) );

	return ret;
}
indesis_type calculate_indesis( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
	using namespace utils::ublas;

	typedef matrix_vector_slice< const vars_type > vars_diagonal_type;
	typedef matrix_vector_slice< const rho1_type > rho1_diagonal_type;

	rho1_diagonal_type rho1_diag = diag( rho1 );
	vars_diagonal_type variances = diag( vars );
	
	return ( element_div( variances*3.0 + rho2 - rho1_diag * 4.0, element_prod(flux,flux) ) ) / 4.5;
}
storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_flux_column( *this, boost::bind(&detail::sample::flux,_1) ),
	m_indices_column( *this, boost::bind(&detail::sample::indices,_1) ),
	m_indesis_column( *this, boost::bind(&detail::sample::indesis,_1) ),
	m_rho1_column( *this, boost::bind(&detail::sample::rho1,_1) ) {
}

namespace filter {
	
flux_threshold::flux_threshold( value_type threshold ):
	m_threshold( threshold ) {
}
flux_threshold::~flux_threshold() {
}
bool flux_threshold::operator() ( const flux_type& flux ) const {
	if( *std::min_element( flux.begin(), flux.end() ) < m_threshold ) {
		throw std::logic_error( (boost::format("Object flux %.2f is under the threshold %.2f") % flux % m_threshold).str() );
		return true;
	}
	return false;
}

}

namespace multiexp {

storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_size_column( *this, boost::bind( &detail::sample::indices_vector_type::size, boost::bind(&detail::sample::indices,_1) ) ) {
}

namespace filter {
	
size_ne::size_ne( const size_type& size ): m_size( size ) {
}
size_ne::~size_ne() {
}
bool size_ne::operator() ( const sample_type& x ) const {
	return x.indices.size() != m_size;
}
	
}
}

}}
