/*
    $Id: integrals.cpp 177 2010-12-13 10:37:30Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <algorithm>
#include <numeric>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <lsp/singular_decomposition.h>

#include "atmos.h"
#include "apply_to_all.h"
#include "atmos.h"
#include "indices.h"
#include "integrals.h"
#include "vector_generator.h"
#include "utils.h"
#include "weif.h"

namespace atmos {
namespace integrals {

const double KM2M = 1000;
const double POWSEE = 0 ;       // Power of Cn2 moment for seeing
const double POWISP = 5.0/3.0;  // Power of Cn2 moment for isoplanatic angle
const double POWEFF = 1.0;      // Power of Cn2 moment for Effective altitude
const double POWISK = 2.0;      // Cn2 second moment power

const double MINRELW = 6e-4 ;   // Minimal ratio of the singular value to the maximal one

static const double powers[] = { POWSEE, POWISP, POWEFF, POWISK };

const unsigned int npower = sizeof(powers)/sizeof(powers[0]);

static const double HBOUND = 0.38;// Height of the boundary layer [kilometers] to fit the Cn2-integral
static const double HTOP = 25.0; // Height of the hight layer [kilometers]

static const int TOPLAYERS =  4; // Number of rejected high-altitude layers
const double JPOW_SEE = 0.6 ;
const double LPOW_SEE = -0.2;
const double LEFF_SEE = 1.1486983549970351;
const double KS = 1.73588e+07;  // Calibration constant in seeing formula (AstL 45, p.399):
                                //\f$ seeing[''] = KS * \lambda_{eff}^{-1/5} M_0^{3/5} *
                                //\sec^{-1}(z)\f$, where \em lambda_eff is in [mkm]*/

const double JPOW_ISP = -0.6;   // Power of Cn-5/3th moment for isoplan. angle
const double LPOW_ISP = 1.2;    // Power of lambda for isoplan. angle
const double LEFF_ISP = 0.43527528164806206;
const double KP = 0.000756348;  // Calibration constant in isoplanatic angle formula (AstL 45, p.399)
                                // \f$ \theta_0[''] = KP * (M_{5/3} /\lambda_{eff}^2)^{-3/5} * \sec^{8/5}(z) \f$, where
                                // \f$ KP = (2.905*(2\pi)^2*(MKM2M)^{-2})^{-3/5}*206265 \f$*/
const double KT = 0.175;        // Atmospheric time constant empirical calibration constant for DESI for 1,3ms
                                // integrations in diameter=2cm apertures with -1km shift

const double SPOW_TC  = -0.6;   // Power of DESI for atm. time constant

functor_factory::dcf_matrix_type functor_factory::calculate( const std::string& spectral_type, const weight_type& awf ) {
	using namespace boost::numeric::ublas;
	using namespace utils::ublas;

	const vector<double> grid( utils::ublas::vector_generator< lg_generator > ( lg_generator(0.0, 0.04, 0.25), 50 ) );

	typedef dcf_matrix_type            result_type;
	typedef dcf_matrix_type::size_type size_type;

	vector<double>::const_iterator lower = std::find_if( grid.begin(), grid.end(), std::bind2nd( std::greater< double >(), HBOUND) );
	vector<double>::const_reverse_iterator higher = std::find_if( grid.rbegin(), grid.rend(), std::bind2nd( std::less< double >(), HTOP) );
	/* FIXME: Weights::grid_type::const_iterator */

	// both layer altitudes and grid are compile-time contants, so it is assertion.
	assert( lower != grid.end() );
	assert( higher != grid.rend() );

	result_type res( npower, 12-2/*dimm*/ );

	vector_range< const vector<double> > h( grid, range( lower.index(), higher.index()+1 ) );
	matrix< double > wfv = vector_of_vector_to_matrix( project( grid, awf) );

	for( unsigned int ipow = 0; ipow < npower; ipow++ ) {
		double moment_power = powers[ipow];
		double weight_power = moment_power / 2.0;
		double scale = pow( KM2M, moment_power );

		vector<double> alt    = utils::ublas::exp( log( h )*moment_power ) * scale;
		vector<double> weight = utils::ublas::exp( log( h )*weight_power );

		matrix<double> A( trans( subrange( wfv, 0, 10, lower.index(), higher.index()+1 ) ) );
		matrix<double> left( identity_matrix<double>( A.size1() ) );
		matrix<double> right( identity_matrix<double>( A.size2() ) );

		for( matrix<double>::size_type i = 0; i < A.size1(); ++i ) {
			alt( i ) /= weight( i );
			row( A, i ) /= weight( i );
		}

		lsp::singular_decomposition< matrix< double > > sd( A );
		sd.apply( left, right );

		matrix_vector_slice< matrix<double> > singular( diag( A ) ); // it is guaranteed that singular values are sorted
		matrix_vector_slice< matrix<double> >::const_iterator small_singular = std::find_if( singular.begin(), singular.end(), std::bind2nd( std::less< double >(), singular(0)*MINRELW) );
		matrix_vector_slice< matrix<double> >::const_iterator it = singular.begin();
		alt = prod( left, alt );
		for( ; it != small_singular; ++it ) {
			row( res, ipow )( it.index() ) = alt( it.index() ) / (*it);
		}
		for( ; it != singular.end(); ++it ) {
			row( res, ipow )( it.index() ) = 0;
		}
		row( res, ipow ) = prod( right, row( res, ipow ) );
	}

	return res;
}
functor_factory::functor_factory() {
}
functor_factory::functor_factory( const functor_factory& factory ) {
	m_storage = factory.m_storage;
}
void functor_factory::swap( functor_factory& factory ) throw() {
	m_storage.swap( factory.m_storage );
}
functor_factory& functor_factory::operator=( const functor_factory& factory ) {
	functor_factory tmp( factory );
	tmp.swap( *this );
	return *this;
}
functor_factory::~functor_factory() {
}
functor_factory::functor_type functor_factory::operator() ( const std::string& spectral_type, const weight_type& awf ) {
	storage_type::const_iterator it = m_storage.find( spectral_type );
	if( it != m_storage.end() )
		return functor_type( it->second );

	std::pair< storage_type::iterator, bool > ret = m_storage.insert( std::make_pair( spectral_type, calculate( spectral_type, awf ) ) );

	return functor_type( ret.first->second );
}

functor::result_type functor::operator() ( const indices_type& indices, const indesis_type& indesis, double mz ) const {
	typedef vector< double >       vector_type;
	typedef vector_type::size_type size_type;
	
	assert( indices.size1() == atmos::channels );
	assert( indices.size2() == atmos::channels );
	assert( indesis.size()  == atmos::channels );

	vector_type aints(10);
	aints(0) = indices(0,0); aints(1) = indices(1,1); aints(2) = indices(2,2); aints(3) = indices(3,3);
	aints(4) = indices(0,1); aints(5) = indices(0,2); aints(6) = indices(0,3);
	aints(7) = indices(1,2); aints(8) = indices(1,3); aints(9) = indices(2,3);

	aints = prod( m_dcf, aints );

	assert( npower == aints.size() );
	for( size_type i = 0; i < npower; ++i ){
		aints(i) = aints(i) / std::pow( mz, 1.0 + powers[i] );
	}

	return result_type( KS * LEFF_SEE * std::pow( aints(0), JPOW_SEE ),
			    KP * LEFF_ISP * std::pow( aints(1), JPOW_ISP ),
			    0.001 * std::pow( aints(1) / aints(2), 1.0/( POWISP-POWEFF ) ),
			    KT * std::pow( indesis(0), SPOW_TC ),
			    aints(0),
			    aints(3) );
}

storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_free_seeing_column( *this, boost::bind(&detail::sample::free_seeing,_1) ),
	m_isoplanatic_angle_column( *this, boost::bind(&detail::sample::isoplanatic_angle,_1) ),
	m_altitude_eff_column( *this, boost::bind(&detail::sample::altitude_eff,_1) ),
	m_time_constant_column( *this, boost::bind(&detail::sample::time_constant,_1) ),
	m_moment_0_column( *this, boost::bind(&detail::sample::moment_0,_1) ),
	m_moment_2_column( *this, boost::bind(&detail::sample::moment_2,_1) ) {
}

namespace filter {
	
bool regular::operator() ( const sample_type& x ) const {
	return ! ( is_regular( x.free_seeing ) &&
		is_regular( x.isoplanatic_angle ) &&
		is_regular( x.altitude_eff ) &&
		is_regular( x.time_constant ) &&
		is_regular( x.moment_0 ) &&
		is_regular( x.moment_2 ) );
}
bool regular::is_regular( double x ) const {
	return ! ( std::isnan( x ) || std::isinf( x ) || ( x < 0 ) );
}

}

}}
