/*
    $Id: longexposure_profile.cpp 171 2010-11-07 18:14:04Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>

#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/function_output_iterator.hpp>
#include <boost/numeric/ublas/banded.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <lsp/cholesky_decomposition.h>
#include <lsp/nnls.h>

#include "apply_to_all.h"
#include "atmos.h"
#include "shortexposure_profile.h"
#include "merge_join.h"
#include "utils.h"

namespace atmos {
namespace shortexposure {
namespace input {

namespace {

// FIXME:
vector_type reorder_flat( const atmos::indices::indices_type& indices ) {
	assert( indices.size1() == atmos::channels );
	assert( indices.size2() == atmos::channels );

	vector_type inds( atmos::indices_size );
	inds(0) = indices(0,0); inds(1) = indices(1,1); inds(2) = indices(2,2); inds(3) = indices(3,3);
	inds(4) = indices(0,1); inds(5) = indices(0,2); inds(6) = indices(0,3);
	inds(7) = indices(1,2); inds(8) = indices(1,3); inds(9) = indices(2,3);
	return inds;
}

struct pack_input {
	typedef void result_type;
	typedef std::pair< atmos::indices::storage::value_type, atmos::profile::storage::value_type > arg_type;

	pack_input( storage& storage ): storage_(storage) { }
	void operator() ( const arg_type& x ) const {
		storage_.push_back( x.first.timestamp, x.second.cn2, reorder_flat(x.first.wind_indices) );
	}
private:
	storage& storage_;
};

}

	
storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_cn2_column(*this, boost::bind(&detail::sample::cn2, _1)),
	m_indices_column(*this, boost::bind(&detail::sample::indices, _1)) {

}
storage::storage( const atmos::indices::storage& indices, const atmos::profile::storage& profile ):
	table_storage< detail::sample >( std::min( indices.size(), profile.size() ) ),
	m_cn2_column(*this, boost::bind(&detail::sample::cn2, _1)),
	m_indices_column(*this, boost::bind(&detail::sample::indices, _1)) {
	using utils::algo::merge_join;

	merge_join( indices.begin(), indices.end(), profile.begin(), profile.end(),
		    boost::make_function_output_iterator( pack_input(*this) ),
		    boost::bind(&atmos::indices::storage::value_type::timestamp, _1),
		    boost::bind(&atmos::profile::storage::value_type::timestamp, _1) );
}

}

namespace profile {

using namespace utils::ublas;

const double LPOW_TAU = 1.2;
const double WPOW_TAU = -0.6;
const double KL = 0.43527528164806206; // 0.5 ** 6/5;
const double KS = 0.057 * 6.3095734448019429e-08 * 1e03;

namespace detail {

//FIXME:
template<class VE> typename VE::size_type nan_elements( const vector_expression<VE>&x ) {
	typename VE::size_type n = 0;
	for( typename VE::size_type j = 0; j < x().size(); j++ )
		n += x()(j) != x()(j) ? 1 : 0;
	return n;
}

}

storage::storage( size_type reserve ):
	table_storage< detail::sample >( reserve ),
	m_windspeed_column(*this, boost::bind(&detail::sample::windspeed, _1)),
	m_chi2_column(*this, boost::bind(&detail::sample::chi2, _1)) {
}

functor::functor( const grid_type& grid, const indices_covariance_type& cov_matrix, double secz, const weight_type& basewf ):
	m_grid( grid ), m_cov_matrix(cov_matrix), m_secz(secz) {
	using namespace utils::ublas;
	using namespace std;

	const value_type cond_numer_threshold = 10000;
	
	banded_matrix< value_type > sigma = sqrt( banded_adaptor< indices_covariance_type >( m_cov_matrix ) );
	for( banded_matrix< value_type >::size_type i = 0; i < sigma.size1(); ++i ) {
		row(m_cov_matrix,i)    /= sigma(i,i);
		column(m_cov_matrix,i) /= sigma(i,i);
	}

	lsp::cholesky_decomposition( m_cov_matrix, upper_tag() );
	triangular_adaptor< indices_covariance_type, lower > R( m_cov_matrix );

	vector_type abs_diag = abs( diag( R ) );
	value_type cond_number = ( detail::nan_elements(abs_diag) ? std::numeric_limits< value_type >::quiet_NaN() : *max_element(abs_diag.begin(),abs_diag.end()) / *min_element(abs_diag.begin(),abs_diag.end()) );

	if( !std::isfinite( cond_number ) || cond_number > cond_numer_threshold ) {
		throw std::logic_error( (boost::format("The condition number of covariance matrix is %1%") % cond_number).str() );
	}

	m_weight_matrix = identity_matrix< value_type >( m_cov_matrix.size1() );
	inplace_solve( R, m_weight_matrix, lower_tag() );
	m_weight_matrix = prod( m_weight_matrix, inv(sigma) );
	R = prod( sigma, R );

	m_problem_matrix = prod( m_weight_matrix, subrange( vector_of_vector_to_matrix( project(grid*secz, basewf) ), 0, m_weight_matrix.size2(),0,grid.size()) );
}
functor::result_type functor::operator() ( const cn2_type& cn2, const indices_type& indices ) const {
	using namespace utils::ublas;

//	const value_type exposure_diff_value = 4e-6 - 1e-6;
	const value_type exposure_diff_value = 6e-6;
	
	vector_type weighted_indices = prod( m_weight_matrix, indices );
	
	vector_type x;
	matrix_type cov = zero_matrix<double>( m_problem_matrix.size2(), m_problem_matrix.size2() );

	lsp::nnls< matrix_type, vector_type > nnls( m_problem_matrix, weighted_indices );
	nnls.solve( x, cov );

	vector_type residual = prod( m_problem_matrix, x ) - weighted_indices;
	value_type chi2 = inner_prod( residual, residual );

	return boost::make_tuple( x * 6.0 / exposure_diff_value / m_secz, chi2 );
}
double functor::tau0( const windspeed_type& cn2wind2 ) const {
	return KS*KL*pow( sum( cn2wind2 ), WPOW_TAU );
}

}
}
}
