/*
    $Id: longexposure_profile.h 171 2010-11-07 18:14:04Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _SHORTEXPOSURE_PROFILE_H
#define _SHORTEXPOSURE_PROFILE_H

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

#include "indices.h"
#include "profile.h"
#include "table_storage.h"

namespace atmos {
namespace shortexposure {
namespace input {
	using namespace boost::numeric::ublas;

	typedef boost::posix_time::ptime time_type;
	typedef double                 value_type;
	typedef vector<value_type>     vector_type;
	typedef vector_type::size_type size_type;
	typedef symmetric_matrix< value_type > symmetric_matrix_type;

	typedef vector_type            indices_type;
	typedef symmetric_matrix_type  indices_covariance_type;
	typedef atmos::profile::cn2_type cn2_type;

namespace detail {
	struct sample {
		typedef atmos::shortexposure::input::time_type    timestamp_type;
		typedef atmos::shortexposure::input::cn2_type     cn2_type;
		typedef atmos::shortexposure::input::indices_type indices_type;

		timestamp_type timestamp;
		cn2_type       cn2;
		indices_type   indices;

		sample( const timestamp_type& timestamp, const cn2_type& cn2, const indices_type& indices ): timestamp(timestamp), cn2(cn2), indices(indices) {
		}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::cn2_type, detail::sample > >      cn2_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::indices_type, detail::sample > >  indices_column_type;
private:
	cn2_column_type      m_cn2_column;
	indices_column_type  m_indices_column;
public:
	storage( size_type reserve = 60 );
	storage( const atmos::indices::storage& indices, const atmos::profile::storage& profile );

	const cn2_column_type&      cn2()      const { return m_cn2_column; }
	const indices_column_type&  indices()  const { return m_indices_column; }

	void push_back( const time_type& timestamp, const cn2_type& cn2, const indices_type& indices ) {
		table_storage< detail::sample >::push_back( detail::sample(timestamp, cn2, indices) );
	}
};
}
namespace profile {
	using namespace boost::numeric::ublas;

	typedef boost::posix_time::ptime time_type;
	typedef double                 value_type;
	typedef vector<value_type>     vector_type;
	typedef matrix<value_type>     matrix_type;

	typedef vector_type            windspeed_type;
	typedef value_type  chi2_type;
	
namespace detail {
	struct sample {
		typedef atmos::shortexposure::profile::time_type      timestamp_type;
		typedef atmos::shortexposure::profile::windspeed_type windspeed_type;
		typedef atmos::shortexposure::profile::chi2_type      chi2_type;

		timestamp_type timestamp;
		windspeed_type windspeed;
		chi2_type      chi2;

		sample( const timestamp_type& timestamp, const windspeed_type& windspeed, const chi2_type& chi2 ):
			timestamp(timestamp), windspeed(windspeed), chi2(chi2) {
		}
	};
}

class storage:
	public table_storage< detail::sample > {
public:
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::windspeed_type, detail::sample > > windspeed_column_type;
	typedef table_storage_column< detail::sample, boost::function1< detail::sample::chi2_type, detail::sample > >      chi2_column_type;
private:
	windspeed_column_type m_windspeed_column;
	chi2_column_type m_chi2_column;
public:
	storage( size_type reserve = 60 );

	const windspeed_column_type& windspeed() const { return m_windspeed_column; }
	const chi2_column_type& chi2() const { return m_chi2_column; }

	void push_back( const time_type& timestamp, const windspeed_type& windspeed, const chi2_type& chi2 ) {
		table_storage< detail::sample >::push_back( detail::sample(timestamp, windspeed, chi2) );
	}
};

class functor {
	public:
		typedef atmos::shortexposure::profile::windspeed_type windspeed_type;
		typedef atmos::shortexposure::profile::value_type     value_type;
		typedef boost::tuple< windspeed_type, value_type > result_type;
		typedef atmos::shortexposure::input::cn2_type     cn2_type;
		typedef atmos::shortexposure::input::indices_type indices_type;
		typedef atmos::shortexposure::input::indices_covariance_type indices_covariance_type;

		typedef vector_type grid_type;
		typedef boost::function1< vector<double>, double > weight_type;
	private:
		grid_type     m_grid;
		indices_covariance_type m_cov_matrix;
		matrix_type   m_weight_matrix;
		matrix_type   m_problem_matrix;
		double        m_secz;

	public:
		functor( const grid_type& grid, const indices_covariance_type& cov_matrix, double secz, const weight_type& basewf );

		result_type operator() ( const cn2_type& cn2, const indices_type& indices ) const;

		double tau0( const windspeed_type& cn2wind2 ) const;
};

}
}
}

#endif // _SHORTEXPOSURE_PROFILE_H
