/*
    $Id: spectral_factory.h 177 2010-12-13 10:37:30Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _SPECTRAL_FACTORY_H
#define _SPECTRAL_FACTORY_H

#include <map>
#include <string>

#include <boost/filesystem/path.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/function.hpp>

#include "nocase_less.h"
#include "vector_generator.h"
#include "weif.h"

namespace spectral {

using namespace boost::numeric::ublas;
using namespace atmos;
typedef vector< double > vector_type;

std::pair< vector_type, vector_type > load_reaction( const boost::filesystem::path& file );
std::pair< vector_type, vector_type > photon_distribution( const vector_type& spectral_lambda, const vector_type& spectral_response, const vector_type& device_lambda, const vector_type& device_response );

void dump_weight( const boost::filesystem::path& file, const boost::function1< vector_type, double >& function, const vector_type& grid = utils::ublas::vector_generator< atmos::lg_generator >( atmos::lg_generator(0.0, 0.04, 0.25), 50 ) );

}

class spectral_factory{
public:
	typedef boost::numeric::ublas::vector< double > vector_type;
	typedef boost::numeric::ublas::vector< boost::numeric::ublas::vector<double> > vector_vector_type;

	typedef utils::ublas::cubic_spline< vector_type, vector_vector_type, vector_type, vector_vector_type > weight_type;
	typedef std::map< std::string, weight_type, utils::algo::nocase_less< std::string > > storage_type;

	typedef storage_type::const_iterator const_iterator;
private:
	boost::filesystem::path m_mass_response;
	boost::filesystem::path m_ccd_response;
	boost::filesystem::path m_wdir;

	utils::ublas::vector_generator< atmos::lg_generator > m_basis_grid;

	storage_type m_nullexposure;
	storage_type m_shortexposure;
	storage_type m_longexposure;

	std::pair< vector_type, vector_type > m_mass_reaction;
	std::pair< vector_type, vector_type > m_dimm_reaction;
private:
	const boost::filesystem::path get_spectrum_filename( const std::string& spectype ) const;
public:
	spectral_factory(const boost::filesystem::path& wd, const boost::filesystem::path& mass_response, const boost::filesystem::path& ccd_response );
	~spectral_factory();

	const weight_type& nullexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam, double dimmbase, double dimmsize );
	const weight_type& shortexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam );
	const weight_type& longexposure_weight( const std::string& spectype, const boost::numeric::ublas::vector<double>& inner_diam, const boost::numeric::ublas::vector<double>& outer_diam );

	const const_iterator nullexposure_begin() const { return m_nullexposure.begin(); }
	const const_iterator nullexposure_end()   const { return m_nullexposure.end(); }
	const const_iterator shortexposure_begin() const { return m_shortexposure.begin(); }
	const const_iterator shortexposure_end()   const { return m_shortexposure.end(); }
	const const_iterator longexposure_begin() const { return m_longexposure.begin(); }
	const const_iterator longexposure_end()   const { return m_longexposure.end(); }
};

#endif //_SPECTRAL_FACTORY_H

