/*
    $Id: weif.h 171 2010-11-07 18:14:04Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _WEIF_H
#define _WEIF_H

#include <boost/function.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include <utility>

#include "spline.h"

namespace atmos {
namespace weight {
	using namespace boost::numeric::ublas;

	typedef double value_type;
	typedef std::complex< value_type > complex_type;
	typedef vector< value_type >   vector_type;
	typedef vector_type::size_type size_type;
	typedef vector_type            weight_type;
	
namespace detail {
	class basic_weight {
	public:
		typedef atmos::weight::value_type value_type;
		typedef value_type  arg_type;
		typedef weight_type result_type;

	protected:
		typedef utils::ublas::cubic_spline< vector_type, vector< complex_type >, vector_type, vector< complex_type > > sfourier_spline_type;
		typedef sfourier_spline_type sfourier_type;
	private:
		value_type do_calculate( value_type height, value_type eps1, value_type eps2, value_type eps3, value_type eps4 ) const;
	public:
		basic_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out, const value_type& freq_power = value_type(0) );

		virtual result_type operator() ( const arg_type& height ) const;
	private:
		vector_type ap_inn_;
		vector_type ap_out_;
		size_type mass_leff_index_;
		value_type mass_leff_;
		sfourier_type mass_sfourier_;
		value_type mass_freq_max_;
		value_type freq_power_;
	};
}

class massdimm_weight:
	public detail::basic_weight {
private:
	value_type do_dimm_calculate( value_type height, value_type diam, value_type dbase, value_type phi ) const;
public:
	massdimm_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& dimm_lambda, const vector_type& dimm_response, const vector_type& ap_inn, const vector_type& ap_out, const value_type& dimmsize, const value_type& dimmbase );
	
	result_type operator() ( const arg_type& height ) const;
private:
	value_type dimmsize_;
	value_type dimmbase_;
	size_type dimm_leff_index_;
	value_type dimm_leff_;
	sfourier_type dimm_sfourier_;
	value_type dimm_freq_max_;
};

class mass_longexposure_weight:
	public detail::basic_weight {
public:
	mass_longexposure_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out );
};

class mass_shortexposure_weight:
	public detail::basic_weight {
public:
	mass_shortexposure_weight( const vector_type& mass_lambda, const vector_type& mass_response, const vector_type& ap_inn, const vector_type& ap_out );
};
}

struct linear_generator {
	typedef double result_type;
	typedef size_t size_type;

	linear_generator( result_type start, result_type step );
	result_type operator()( size_type i ) const;
	size_type size( result_type max ) const;
private:
	result_type start_,step_;
};

struct lg_generator {
	typedef double result_type;
	typedef size_t size_type;

	lg_generator( result_type start, result_type lstep, result_type gstep );
	result_type operator()( size_type i ) const;
	size_type size( result_type max ) const;
private:
	result_type start_,t_,q_;
	size_type max_;
};

}

#endif // _WEIF_H
