/*
    $Id: worksheet.h 166 2010-10-29 15:22:58Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _WORKSHEET_H
#define _WORKSHEET_H

#include "background_factory.h" // for background_factory
#include "file_input.h"       // for file_input
#include "file_output.h"      // for file_output
#include "params.h"           // for parameters_*
#include "spectral_factory.h" // for spectral_factory
#include "typesdef.h"         // for nocase_less

#include "indices.h"
#include "integrals.h"
#include "profile.h"

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/program_options.hpp>

template<class Context> class local_context {
	public:
		typedef Context                context_type;
		typedef context_type&          context_reference_type;
		typedef const context_type&    context_const_reference_type;
		typedef local_context<Context> local_context_type;
	private:
		context_reference_type m_ctx;
	protected:
		inline context_reference_type       context()       throw() { return m_ctx; }
		inline context_const_reference_type context() const throw() { return m_ctx; }
	public:
		local_context( context_reference_type ctx ) throw() : m_ctx(ctx) {}
};

class worksheet {
	friend class local_context<worksheet>;
	public:
		typedef local_context<worksheet> context_type;
		typedef parameters_container< std::map< std::string, std::string, utils::algo::nocase_less< std::string > > > parameters_container_type;
		typedef parameters_proxy< parameters_container_type > parameters_proxy_type;
		typedef spectral_factory spectral_factory_type;
		typedef background_factory background_factory_type;
		typedef atmos::integrals::functor_factory integrals_factory_type;
	public:
		static int parse_program_options( int argc, char** argv, boost::program_options::variables_map& va );
	public:
		struct profile_grid;
	private:
		class input:
			public file_input,
			public local_context< worksheet > {
		private:
			class measurement_accumulator: public local_context< worksheet > {
			private:
				boost::posix_time::ptime m_starttime;
			public:
				measurement_accumulator( const boost::posix_time::ptime& time, context_reference_type ctx ) throw() : local_context_type( ctx ), m_starttime(time) {};
				virtual ~measurement_accumulator() = 0;
				virtual void accumulate( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) = 0;
				virtual void dimm_accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) = 0;
				virtual void multiexp_accumulate( const time_type& time, const flux_type& mean, const vars_type& vars ) = 0;
				virtual void flush() = 0;

				template<class T> bool type() const {
					return dynamic_cast<const T*>(this);
				}

				inline const boost::posix_time::ptime& starttime() const { return m_starttime; }
			};

			class dummy_accumulator: public measurement_accumulator {
			public:
				dummy_accumulator( const boost::posix_time::ptime& time, context_reference_type ctx ) throw(): measurement_accumulator( time, ctx ) {};
				virtual ~dummy_accumulator() {};
				virtual void accumulate(  const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {}
				virtual void dimm_accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {}
				virtual void multiexp_accumulate( const time_type& time, const flux_type& mean, const vars_type& vars ) {}
				virtual void flush() {}
			};

			class background_accumulator;
			class normal_accumulator;
		private:
			std::auto_ptr<measurement_accumulator> m_maccum;

			template<class T> void switch_accumulator( const boost::posix_time::ptime& time ) {
				std::auto_ptr< measurement_accumulator > new_accumulator( new T( time,context() ) );
				std::swap( m_maccum, new_accumulator );
				new_accumulator->flush();
			}

			template<class T> bool accumulator_type() const {
				return m_maccum->type<T>();
			}
		public:
			void on_comment( const std::string& line );
			void on_parameter( const time_type& time, const std::string& key, const std::string& value );
			void on_object( const time_type& time, const object_type& obj );
			void on_background( const time_type& time );
			void on_normal( const time_type& time );
			void on_flux( const time_type& time );
			void on_measurements( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 );
			void on_dimm( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars );
			void on_multiexp( const time_type& time, const flux_type& mean, const vars_type& vars );
		public:
			input( std::istream& stm, context_reference_type ctx ): file_input( stm ), local_context_type( ctx ), m_maccum( new dummy_accumulator( time_type() ,ctx) ) {}
		};
		
		class object;
	private:
		const boost::program_options::variables_map& m_va;
		
		parameters_container_type m_pcontainer;
		parameters_proxy_type     m_pproxy;

		spectral_factory_type     m_sfactory;

		background_factory_type   m_bfactory;

		integrals_factory_type    m_ifactory;
		
		input       m_file_input;
		file_output m_file_output;

		std::auto_ptr<object> m_curobject;
	private:
		background_model* select_background_model( const boost::posix_time::ptime& time ) const;
	public:
		worksheet( const boost::program_options::variables_map& va );
		~worksheet( );

		inline const parameters_proxy_type& pproxy()     const { return m_pproxy;   }
		inline const spectral_factory_type& sfactory()   const { return m_sfactory; }
		inline const background_factory_type& bfactory() const { return m_bfactory; }
		inline const integrals_factory_type& ifactory()  const { return m_ifactory; }

		double sun_altitude( const boost::posix_time::ptime& time ) const;

		inline object& curobject() {
			if( ! m_curobject.get() )
				throw std::logic_error("There is not any object");
			return *m_curobject;
		}

		void pre();
		void run();
		void post();
};

#endif // _WORKSHEET_H
